use axum::{
    routing::{get, post},
    Router,
};
use serde::Serialize;
use std::{net::SocketAddr, sync::Arc};

#[async_trait::async_trait]
pub trait Handler {
    type Ok: Serialize;
    type Err: Serialize;

    async fn handle(self, ctx: &Context) -> Result<Self::Ok, Self::Err>;
}

mod balance;
mod deposit;
mod get;
mod lock;
mod subscribe;
mod tx;
mod update;

#[derive(Clone)]
pub struct Context {
    ledger: Arc<verax::Ledger<verax::storage::Cache<verax::storage::SQLite>>>,
}

#[tokio::main]
async fn main() -> std::io::Result<()> {
    if std::env::var_os("RUST_LOG").is_none() {
        std::env::set_var("RUST_LOG", "actix_web=info");
    }
    env_logger::init_from_env(env_logger::Env::new().default_filter_or("info"));

    let settings = "sqlite:./ledger.db"
        .parse::<verax::storage::sqlite::SqliteConnectOptions>()
        .expect("valid settings")
        .journal_mode(verax::storage::sqlite::SqliteJournalMode::Wal)
        .create_if_missing(true);

    let pool = verax::storage::sqlite::SqlitePoolOptions::new()
        .connect_with(settings)
        .await
        .expect("pool");

    let storage = verax::storage::SQLite::new(pool.clone());
    storage.setup().await.expect("setup");

    let storage = verax::storage::SQLite::new(pool.clone());
    let storage = verax::storage::Cache::new(storage);
    let ledger = verax::Ledger::new(storage.into());

    let app = Router::new()
        .route("/balance/:account", get(balance::handler))
        .route("/deposit", post(deposit::handler))
        .route("/lock", post(lock::handler))
        .route("/subscribe", post(subscribe::handler))
        .route("/tx", post(tx::handler))
        .route("/:id", post(update::handler))
        .route("/:id", get(get::handler))
        .with_state(Context { ledger });

    let addr = SocketAddr::from(([127, 0, 0, 1], 8080));
    let listener = tokio::net::TcpListener::bind(addr).await.unwrap();
    axum::serve(listener, app).await?;
    Ok(())
}
