use std::collections::HashMap;

use super::{
    expr::{Expr, ExprOp},
    opcode::OpCode,
    parser::parse_query,
    runtime::execute,
    value::Value,
    Addr, Error, Register,
};
use crate::Transaction;

#[derive(Debug, Clone)]
pub struct Program {
    /// Debug op-codes, with the unresolved labels.
    dbg_opcodes: Vec<OpCode>,
    /// The list of opcodes that make up the program
    opcodes: Vec<OpCode>,
    /// If the program has some boilerplate that can be skipped. This is non-zero when the initial
    /// program has been executed at compile time, and the register has been populated to
    /// `initial_register`. Everytime a new Runtime is created, instead of executing the program,
    /// it can safely start-off a different Address, and clone the initial_register.
    start_at: Addr,
    /// The state of the register
    initial_register: Vec<Value>,
}

pub struct Compiler<'a> {
    expr: &'a Expr,
    current_register: Register,
    labels: Addr,
}

impl<'a> Compiler<'a> {
    pub fn new(expr: &'a Expr) -> Self {
        Self {
            expr,
            current_register: 0,
            labels: 0,
        }
    }

    fn next_label(&mut self) -> Addr {
        self.labels += 1;
        self.labels
    }

    fn next_register(&mut self) -> usize {
        self.current_register += 1;
        self.current_register
    }

    fn compile_expr(
        &mut self,
        expr: &'a Expr,
        exit_label: Addr,
    ) -> Result<(Vec<OpCode>, usize), Error> {
        let mut return_value = self.next_register();
        Ok(match expr {
            Expr::Variable(name) => (
                vec![OpCode::LOAD_EXTERNAL(return_value, name.into())],
                return_value,
            ),
            Expr::String(string) => (
                vec![OpCode::LOAD(return_value, Value::String(string.clone()))],
                return_value,
            ),
            Expr::Number(number) => (
                vec![OpCode::LOAD(return_value, Value::Number(*number))],
                return_value,
            ),
            Expr::Op { op, terms } => {
                let mut opcodes = vec![];
                let current_exit_label = self.next_label();

                let mut compiled_terms = terms
                    .iter()
                    .map(|term| self.compile_expr(term, current_exit_label))
                    .collect::<Result<Vec<_>, _>>()?;

                match op {
                    ExprOp::Not => {
                        let (mut opcodes, term_return_value) =
                            compiled_terms.pop().ok_or(Error::EmptyRegisters)?;
                        if !compiled_terms.is_empty() {
                            return Err(Error::UnexpectedExprState);
                        }
                        opcodes.push(OpCode::NOT(return_value, return_value));
                        return_value = term_return_value;
                    }
                    ExprOp::Eq => {
                        let cmp = self.next_register();
                        let last_value = self.next_register();
                        let mut iter = compiled_terms.into_iter();
                        let (mut term_opcodes, return_from_expr) =
                            iter.next().ok_or(Error::MissingNextRule)?;

                        opcodes.append(&mut term_opcodes);
                        opcodes.push(OpCode::MOV(last_value, return_from_expr));

                        for (mut term_opcodes, term_return) in iter {
                            opcodes.append(&mut term_opcodes);
                            opcodes.push(OpCode::EQ(cmp, last_value, term_return));
                            opcodes.push(OpCode::JNE(exit_label, cmp));
                            opcodes.push(OpCode::MOV(last_value, term_return));
                        }

                        opcodes.push(OpCode::LOAD(return_value, true.into()));
                    }
                    ExprOp::Add | ExprOp::Mul | ExprOp::Div | ExprOp::Sub => {
                        let mut iter = compiled_terms.into_iter();
                        let (mut term_opcodes, return_value) =
                            iter.next().ok_or(Error::MissingNextRule)?;
                        opcodes.append(&mut term_opcodes);

                        for (mut term_opcodes, term_return) in iter {
                            opcodes.append(&mut term_opcodes);
                            opcodes.push(match op {
                                ExprOp::Add => OpCode::ADD(return_value, return_value, term_return),
                                ExprOp::Mul => OpCode::MUL(return_value, return_value, term_return),
                                ExprOp::Div => OpCode::DIV(return_value, return_value, term_return),
                                ExprOp::Sub => OpCode::SUB(return_value, return_value, term_return),
                                _ => unreachable!(),
                            });
                        }

                        return Ok((opcodes, return_value));
                    }
                    ExprOp::Or | ExprOp::And => {
                        opcodes = vec![OpCode::LOAD(return_value, Value::Bool(false))];
                        for (mut term_opcodes, term_return_value) in compiled_terms.into_iter() {
                            opcodes.append(&mut term_opcodes);
                            opcodes.push(OpCode::MOV(return_value, term_return_value));
                            match op {
                                ExprOp::Or => opcodes.push(OpCode::JEQ(exit_label, return_value)),
                                ExprOp::And => opcodes.push(OpCode::JNE(exit_label, return_value)),
                                _ => unreachable!(),
                            };
                        }
                    }
                    x => panic!("unimplemented {:#?}", x),
                }

                opcodes.push(OpCode::LABEL(current_exit_label));
                (opcodes, return_value)
            }
        })
    }

    fn resolve_label_to_addr(opcodes: Vec<OpCode>) -> Result<Vec<OpCode>, Error> {
        let mut pos = 0;
        let used_labels = opcodes
            .iter()
            .map(|x| match x {
                OpCode::LABEL(label) => (x, pos),
                _ => {
                    pos = pos + 1;
                    (x, pos - 1)
                }
            })
            .filter_map(|(opcode, pos)| match opcode {
                OpCode::LABEL(id) => Some((*id, pos)),
                _ => None,
            })
            .collect::<HashMap<_, _>>();

        opcodes
            .into_iter()
            .filter(|opcode| !matches!(opcode, OpCode::LABEL(_)))
            .map(|opcode| {
                Ok(match opcode {
                    OpCode::JMP(label) => {
                        OpCode::JMP(*used_labels.get(&label).ok_or(Error::UnknownLabel(label))?)
                    }
                    OpCode::JEQ(register, label) => OpCode::JEQ(
                        register,
                        *used_labels.get(&label).ok_or(Error::UnknownLabel(label))?,
                    ),
                    OpCode::JNE(register, label) => OpCode::JNE(
                        register,
                        *used_labels.get(&label).ok_or(Error::UnknownLabel(label))?,
                    ),
                    x => x,
                })
            })
            .collect()
    }

    pub fn compile(mut self) -> Result<Vec<OpCode>, Error> {
        let program_exit = self.next_label();
        let (mut opcodes, return_value) = self.compile_expr(self.expr, program_exit)?;
        opcodes.push(OpCode::LABEL(program_exit));
        opcodes.push(OpCode::HLT(return_value));
        Ok(opcodes)
    }
}

impl Program {
    pub fn new(code: &str) -> Result<Self, Error> {
        let ast = parse_query(code)?;

        let opcodes = ast.where_clause.map_or_else(
            || Ok(vec![OpCode::LOAD(0, true.into()), OpCode::HLT(0)]),
            |expr| Compiler::new(&expr).compile(),
        )?;

        println!(
            "{}",
            opcodes
                .iter()
                .map(|x| match x {
                    OpCode::HLT(_) | OpCode::LABEL(_) => x.to_string(),
                    x => format!("\t{}", x.to_string()),
                })
                .collect::<Vec<_>>()
                .join("\n")
        );

        Ok(Self {
            dbg_opcodes: opcodes.clone(),
            opcodes: Compiler::resolve_label_to_addr(opcodes)?,
            start_at: 0,
            initial_register: vec![],
        })
    }

    /// Returns a human readable version of the compiled program (generated op-codes)
    pub fn debug(&self) -> String {
        self.dbg_opcodes
            .iter()
            .map(|x| match x {
                OpCode::HLT(_) | OpCode::LABEL(_) => x.to_string(),
                x => format!("\t{}", x.to_string()),
            })
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn dump(&self) -> String {
        self.opcodes
            .iter()
            .enumerate()
            .map(|(pos, opcode)| format!("{}: {}", pos, opcode.to_string()))
            .collect::<Vec<_>>()
            .join("\n")
    }

    /// Returns the opcodes of the program expression and the usize of the register where the result
    /// Creates a program tailored for a transaction
    pub fn execute(&self, transaction: Option<&Transaction>) -> Result<Value, Error> {
        execute(
            transaction,
            &self.opcodes,
            self.initial_register.clone(),
            self.start_at,
        )
    }
}

#[cfg(test)]
mod test {
    use super::Program;
    use crate::filter_expr::{opcode::OpCode, value::Value};

    #[test]
    fn parse() {
        let x = Program::new(
            r#"
            WHERE
                $foo = 3
                AND $bar = "bar"
                AND ($a = $b OR $b = $c)
                AND $foo = 1 + 2 + 3 +  ((1+ 2-3*4-5) * $bar.tx.lol)
            ORDER BY $bar DESC
        "#,
        )
        .unwrap();
        panic!("{}\n\n{}", x.debug(), x.dump());
    }

    #[test]
    fn simple_program() {
        let program = Program {
            dbg_opcodes: vec![],
            opcodes: vec![
                OpCode::LOAD(1, 12.into()),
                OpCode::LOAD(2, 13.into()),
                OpCode::ADD(3, 1, 2),
                OpCode::ADD(4, 0, 3),
                OpCode::HLT(4),
            ],
            start_at: 0,
            initial_register: vec![15.into()],
        };
        let x = program.execute(None).expect("valid execution");
        assert_eq!(x, 40.into());
    }
}
