use super::{opcode::OpCode, value::Value, Addr, Error};
use crate::Transaction;
use num::CheckedAdd;
use std::{collections::VecDeque, ops::Deref};

#[derive(Debug, PartialEq, PartialOrd)]
/// Value or reference to a value.
///
/// A reference to a value is being used to avoid cloning from the source code to the registers,
/// instead to just use a readonly reference
enum ValueOrRef<'a> {
    Ref(&'a Value),
    Value(Value),
}

impl<'a> Clone for ValueOrRef<'a> {
    fn clone(&self) -> Self {
        match self {
            ValueOrRef::Ref(value) => ValueOrRef::Value((*value).clone()),
            ValueOrRef::Value(value) => ValueOrRef::Value(value.clone()),
        }
    }
}

impl<'a> Into<Value> for ValueOrRef<'a> {
    fn into(self) -> Value {
        match self {
            ValueOrRef::Ref(value) => value.clone(),
            ValueOrRef::Value(value) => value,
        }
    }
}

impl Deref for ValueOrRef<'_> {
    type Target = Value;

    fn deref(&self) -> &Self::Target {
        match self {
            ValueOrRef::Ref(value) => *value,
            ValueOrRef::Value(value) => &value,
        }
    }
}

impl<'a> From<&'a Value> for ValueOrRef<'a> {
    fn from(value: &'a Value) -> Self {
        ValueOrRef::Ref(value)
    }
}

impl<'a> From<Value> for ValueOrRef<'a> {
    fn from(value: Value) -> Self {
        ValueOrRef::Value(value)
    }
}

macro_rules! get {
    ($r:expr,$pos:expr) => {
        ($r.get($pos)
            .ok_or_else(|| Error::RegisterOutOfBoundaries($pos))?)
    };
}

macro_rules! set {
    ($r:expr, $pos:expr, $new_value:expr) => {
        (if let Some(previous_value) = $r.get_mut($pos) {
            *previous_value = $new_value;
        } else {
            $r.push_back($new_value);
            if $r.len() != $pos + 1 {
                return Err(Error::RegisterOutOfBoundaries($pos));
            }
        })
    };
}

#[inline]
pub fn execute(
    transaction: Option<&Transaction>,
    code: &[OpCode],
    initial_registers: Vec<Value>,
    start_at: Addr,
) -> Result<Value, Error> {
    let mut execution = start_at;
    let mut registers: VecDeque<ValueOrRef> = initial_registers
        .iter()
        .map(|a| a.into())
        .collect::<VecDeque<_>>();

    loop {
        match code.get(execution).ok_or(Error::OutOfBoundaries)? {
            OpCode::LOAD(dst, ref val) => set!(registers, *dst, val.into()),
            OpCode::CPY(dst, reg2) => {
                let value = get!(registers, *reg2).clone();
                set!(registers, *dst, value);
            }
            OpCode::MOV(dst, reg2) => {
                let _ = get!(registers, *reg2);
                let previous_value = std::mem::replace(&mut registers[*reg2], Value::Nil.into());
                set!(registers, *dst, previous_value);
            }
            OpCode::ADD(dst, a, b) => {
                let new_value = get!(registers, *a)
                    .checked_add(get!(registers, *b))
                    .ok_or(Error::Overflow)?
                    .into();
                set!(registers, *dst, new_value);
            }
            OpCode::AND(dst, reg2, reg3) => {
                let new_value = Value::Bool(
                    get!(registers, *reg2).as_boolean()? && get!(registers, *reg3).as_boolean()?,
                )
                .into();
                set!(registers, *dst, new_value);
            }
            OpCode::OR(dst1, reg2, reg3) => {
                todo!()
            }
            OpCode::XOR(dst, reg2, reg3) => {
                todo!()
            }
            OpCode::NOT(dst, reg2) => {
                let new_value = Value::Bool(!get!(registers, *reg2).as_boolean()?).into();
                set!(registers, *dst, new_value);
            }
            OpCode::JMP(addr) => {
                execution = *addr;
                continue;
            }
            OpCode::JEQ(reg, addr) => {
                if get!(registers, *reg).as_boolean()? {
                    execution = *addr;
                    continue;
                }
            }
            OpCode::JNE(reg, addr) => {
                if !get!(registers, *reg).as_boolean()? {
                    execution = *addr;
                    continue;
                }
            }
            OpCode::HLT(return_register) => {
                return registers
                    .remove(*return_register)
                    .map(|x| x.into())
                    .ok_or(Error::EmptyRegisters)
            }
            _ => todo!(),
        }

        execution += 1;
    }
}
