use crate::{
    filter_expr::Error, payment::PaymentTo, AccountId, Amount, Asset, PaymentFrom, RevId, Status,
    Tag, TxId, Type,
};
use chrono::{DateTime, Utc};
use num::CheckedAdd;
use std::ops::Add;

use super::expr::{Expr, Variable};

#[derive(Clone, Debug, PartialEq, PartialOrd)]
pub enum Value {
    Nil,
    Amount(Amount),
    Asset(Asset),
    AccountId(AccountId),
    TxId(TxId),
    Type(Type),
    RevId(RevId),
    Status(Status),
    Tag(Tag),
    Date(DateTime<Utc>),
    String(String),
    Number(i128),
    From(Vec<PaymentFrom>),
    To(Vec<PaymentTo>),
    Tags(Vec<Tag>),
    Bool(bool),
    Variable(Vec<String>),
}

impl From<&Variable> for Value {
    fn from(value: &Variable) -> Self {
        Self::Variable(value.to_vec())
    }
}

impl From<bool> for Value {
    fn from(x: bool) -> Self {
        Value::Bool(x)
    }
}

impl From<i128> for Value {
    fn from(x: i128) -> Self {
        Value::Number(x)
    }
}

impl Value {
    /// Attempts to convert the value as a boolean if possible or returns an error.
    pub fn as_boolean(&self) -> Result<bool, Error> {
        match self {
            Value::Bool(b) => Ok(*b),
            Value::Nil => Ok(false),
            Value::Number(x) => Ok(*x > 0),
            _ => panic!("Cannot convert to bool"),
        }
    }

    pub fn as_number(&self) -> Result<i128, Error> {
        match self {
            Value::Number(x) => Ok(*x),
            Value::String(x) => x.parse().map_err(|_| Error::NotANumber),
            _ => Err(Error::NotANumber),
        }
    }
}

impl Add for Value {
    type Output = Self;

    fn add(self, other: Self) -> Self {
        Value::Number(
            self.as_number()
                .expect("Not a number")
                .checked_add(other.as_number().expect("Not a number"))
                .expect("Overflow"),
        )
    }
}

impl CheckedAdd for Value {
    fn checked_add(&self, other: &Self) -> Option<Self> {
        self.as_number()
            .ok()?
            .checked_add(other.as_number().ok()?)
            .map(Value::Number)
    }
}
